@@ -36,25 +36,10 @@
 #include <typeinfo>
 #include <algorithm>
 #include <ostream>
+#include <utility>
 
 namespace nghttp2 {
 
-#if __cplusplus > 201103L
-using std::make_unique;
-#else  // __cplusplus <= 201103L
-template <typename T, typename... U>
-typename std::enable_if<!std::is_array<T>::value, std::unique_ptr<T>>::type
-make_unique(U &&... u) {
-  return std::unique_ptr<T>(new T(std::forward<U>(u)...));
-}
-
-template <typename T>
-typename std::enable_if<std::is_array<T>::value, std::unique_ptr<T>>::type
-make_unique(size_t size) {
-  return std::unique_ptr<T>(new typename std::remove_extent<T>::type[size]());
-}
-#endif // __cplusplus <= 201103L
-
 // std::forward is constexpr since C++14
 template <typename... T>
 constexpr std::array<
@@ -105,21 +90,17 @@ template <typename T> struct DList {
   DList &operator=(const DList &) = delete;
 
   DList(DList &&other) noexcept
-      : head(other.head), tail(other.tail), len(other.len) {
-    other.head = other.tail = nullptr;
-    other.len = 0;
-  }
+      : head{std::exchange(other.head, nullptr)},
+        tail{std::exchange(other.tail, nullptr)},
+        len{std::exchange(other.len, 0)} {}
 
   DList &operator=(DList &&other) noexcept {
     if (this == &other) {
       return *this;
     }
-    head = other.head;
-    tail = other.tail;
-    len = other.len;
-
-    other.head = other.tail = nullptr;
-    other.len = 0;
+    head = std::exchange(other.head, nullptr);
+    tail = std::exchange(other.tail, nullptr);
+    len = std::exchange(other.len, 0);
 
     return *this;
   }
@@ -201,7 +182,7 @@ constexpr double operator"" _ms(unsigned long long ms) { return ms / 1000.; }
 // Returns a copy of NULL-terminated string [first, last).
 template <typename InputIt>
 std::unique_ptr<char[]> strcopy(InputIt first, InputIt last) {
-  auto res = make_unique<char[]>(last - first + 1);
+  auto res = std::make_unique<char[]>(last - first + 1);
   *std::copy(first, last, res.get()) = '\0';
   return res;
 }
@@ -266,10 +247,7 @@ public:
   ImmutableString(const ImmutableString &other)
       : len(other.len), base(copystr(std::begin(other), std::end(other))) {}
   ImmutableString(ImmutableString &&other) noexcept
-      : len(other.len), base(other.base) {
-    other.len = 0;
-    other.base = "";
-  }
+      : len{std::exchange(other.len, 0)}, base{std::exchange(other.base, "")} {}
   ~ImmutableString() {
     if (len) {
       delete[] base;
@@ -294,10 +272,8 @@ public:
     if (len) {
       delete[] base;
     }
-    len = other.len;
-    base = other.base;
-    other.len = 0;
-    other.base = "";
+    len = std::exchange(other.len, 0);
+    base = std::exchange(other.base, "");
     return *this;
   }
 
@@ -417,7 +393,7 @@ public:
   explicit StringRef(const char *s) : base(s), len(strlen(s)) {}
   constexpr StringRef(const char *s, size_t n) : base(s), len(n) {}
   template <typename CharT>
-  StringRef(const CharT *s, size_t n)
+  constexpr StringRef(const CharT *s, size_t n)
       : base(reinterpret_cast<const char *>(s)), len(n) {}
   template <typename InputIt>
   StringRef(InputIt first, InputIt last)
